/*
 * Copyright (C) 2005 - 2011 Emulex
 * All rights reserved.
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License version 2
 * as published by the Free Software Foundation. The full GNU General
 * Public License is included in this distribution in the file called COPYING.
 *
 * Contact Information:
 * linux-drivers@emulex.com
 *
 * Emulex
 * 3333 Susan Street
 * Costa Mesa, CA 92626
 */

#include "be.h"

#ifndef PCIE_SET_READRQ_defined
int pcie_set_readrq(struct pci_dev *dev, int rq)
{
	int cap, err = -EINVAL;
	u16 ctl, v;

	if (rq < 128 || rq > 4096 || !is_power_of_2(rq))
		goto out;

	v = (ffs(rq) - 8) << 12;

	cap = pci_find_capability(dev, PCI_CAP_ID_EXP);
	if (!cap)
		goto out;

	err = pci_read_config_word(dev, cap + PCI_EXP_DEVCTL, &ctl);
	if (err)
		goto out;

	if ((ctl & PCI_EXP_DEVCTL_READRQ) != v) {
		ctl &= ~PCI_EXP_DEVCTL_READRQ;
		ctl |= v;
		err = pci_write_config_dword(dev, cap + PCI_EXP_DEVCTL, ctl);
	}

out:
	return err;
}
#endif /* PCIE_SET_READRQ_defined */

#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 28)
void be_netdev_ops_init(struct net_device *netdev, struct net_device_ops *ops)
{
	netdev->open = ops->ndo_open;
	netdev->stop = ops->ndo_stop;
	netdev->hard_start_xmit = ops->ndo_start_xmit;
	netdev->set_mac_address = ops->ndo_set_mac_address;
	netdev->get_stats = ops->ndo_get_stats;
	netdev->set_multicast_list = ops->ndo_set_rx_mode;
	netdev->change_mtu = ops->ndo_change_mtu;
	netdev->vlan_rx_register = ops->ndo_vlan_rx_register;
	netdev->vlan_rx_add_vid = ops->ndo_vlan_rx_add_vid;
	netdev->vlan_rx_kill_vid = ops->ndo_vlan_rx_kill_vid;
	netdev->do_ioctl = ops->ndo_do_ioctl;
#ifdef CONFIG_NET_POLL_CONTROLLER
	netdev->poll_controller = ops->ndo_poll_controller;
#endif
#if LINUX_VERSION_CODE >= KERNEL_VERSION(2, 6, 27)
	netdev->select_queue = ops->ndo_select_queue;
#endif
}
#endif

/* New NAPI backport */
#if LINUX_VERSION_CODE < KERNEL_VERSION(2, 6, 27)

int be_poll_compat(struct net_device *netdev, int *budget)
{
	struct napi_struct *napi = netdev->priv;
	u32 work_done, can_do;

	can_do = min(*budget, netdev->quota);
	work_done = napi->poll(napi, can_do);

	*budget -= work_done;
	netdev->quota -= work_done;
	return (work_done >= can_do);
}

void netif_napi_add_compat(struct net_device *netdev,
		struct napi_struct *napi,
		int (*poll) (struct napi_struct *, int), int weight)
{
	struct net_device *nd;

	nd = alloc_netdev(0, "", ether_setup);
	if (!nd)
		return;
	nd->priv = napi;
	nd->weight = BE_NAPI_WEIGHT;
	nd->poll = be_poll_compat;
	set_bit(__LINK_STATE_START, &nd->state);
	napi->poll = poll;
	napi->dev = nd;
#ifdef RHEL
	napi->napi.dev = netdev;
#endif
}

void netif_napi_del_compat(struct net_device *netdev)
{
	struct be_adapter *adapter = netdev_priv(netdev);
	struct napi_struct *napi;
	struct be_eq_obj *eqo;
	int i;

	for_all_evt_queues(adapter, eqo, i) {
		napi = &eqo->napi;
		if (napi->dev) {
			free_netdev(napi->dev);
			napi->dev = NULL;
		}
	}
}
#endif /* New NAPI backport */

#ifndef TX_MQ
struct net_device *alloc_etherdev_mq_compat(int sizeof_priv,
					unsigned int queue_count)
{
	return alloc_etherdev(sizeof_priv);
}

void netif_wake_subqueue_compat(struct net_device *dev, u16 queue_index)
{
	netif_wake_queue(dev);
}

void netif_stop_subqueue_compat(struct net_device *dev, u16 queue_index)
{
	netif_stop_queue(dev);
}

int __netif_subqueue_stopped_compat(const struct net_device *dev,
						u16 queue_index)
{
	return netif_queue_stopped(dev);
}

u16 skb_get_queue_mapping_compat(const struct sk_buff *skb)
{
        return 0;
}

void netif_set_real_num_tx_queues_compat(struct net_device *dev,
					unsigned int txq)
{
	return;
}

u16 skb_tx_hash_compat(const struct net_device *dev,
			const struct sk_buff *skb)
{
	return 0;
}
#endif
