/************** Backport of INET_LRO for SLES 10 *****************************/
#ifndef GRO_defined
#define lro_flush_all			lro_flush_all_compat
#define lro_vlan_hwaccel_receive_frags 	lro_vlan_hwaccel_receive_frags_compat
#define lro_receive_frags		lro_receive_frags_compat

struct net_lro_stats {
	unsigned long aggregated;
	unsigned long flushed;
	unsigned long no_desc;
};

struct net_lro_desc {
	struct sk_buff *parent;
	struct sk_buff *last_skb;
	struct skb_frag_struct *next_frag;
	struct iphdr *iph;
	struct tcphdr *tcph;
	struct vlan_group *vgrp;
	__wsum  data_csum;
	u32 tcp_rcv_tsecr;
	u32 tcp_rcv_tsval;
	u32 tcp_ack;
	u32 tcp_next_seq;
	u32 skb_tot_frags_len;
	u32 ack_cnt;
	u16 ip_tot_len;
	u16 tcp_saw_tstamp;		/* timestamps enabled */
	u16 tcp_window;
	u16 vlan_tag;
	int pkt_aggr_cnt;		/* counts aggregated packets */
	int vlan_packet;
	int mss;
	int active;
};

struct net_lro_mgr {
	struct net_device *dev;
	struct net_lro_stats stats;

	/* LRO features */
	unsigned long features;
#define LRO_F_NAPI            1  /* Pass packets to stack via NAPI */
#define LRO_F_EXTRACT_VLAN_ID 2  /* Set flag if VLAN IDs are extracted
				from received packets and eth protocol
				    is still ETH_P_8021Q */

	u32 ip_summed;      /* Set in non generated SKBs in page mode */
	u32 ip_summed_aggr; /* Set in aggregated SKBs: CHECKSUM_UNNECESSARY
			     * or CHECKSUM_NONE */

	int max_desc; /* Max number of LRO descriptors  */
	int max_aggr; /* Max number of LRO packets to be aggregated */

	struct net_lro_desc *lro_arr; /* Array of LRO descriptors */

	/* Optimized driver functions
	 * get_skb_header: returns tcp and ip header for packet in SKB
	 */
	int (*get_skb_header)(struct sk_buff *skb, void **ip_hdr,
			      void **tcpudp_hdr, u64 *hdr_flags, void *priv);

	/* hdr_flags: */
#define LRO_IPV4 1 /* ip_hdr is IPv4 header */
#define LRO_TCP  2 /* tcpudp_hdr is TCP header */

	/*
	 * get_frag_header: returns mac, tcp and ip header for packet in SKB
	 *
	 * @hdr_flags: Indicate what kind of LRO has to be done
	 *             (IPv4/IPv6/TCP/UDP)
	 */
	int (*get_frag_header)(struct skb_frag_struct *frag, void **mac_hdr,
			       void **ip_hdr, void **tcpudp_hdr, u64 *hdr_flags,
			       void *priv);
};

extern void lro_receive_skb(struct net_lro_mgr *lro_mgr, struct sk_buff *skb,
			void *priv);

extern void lro_vlan_hwaccel_receive_skb(struct net_lro_mgr *lro_mgr,
			struct sk_buff *skb, struct vlan_group *vgrp,
			u16 vlan_tag, void *priv);

/* This functions aggregate fragments and generate SKBs do pass
 * the packets to the stack.
 *
 * @lro_mgr: LRO manager to use
 * @frags: Fragment to be processed. Must contain entire header in first
 *         element.
 * @len: Length of received data
 * @true_size: Actual size of memory the fragment is consuming
 * @priv: Private data that may be used by driver functions
 *        (for example get_tcp_ip_hdr)
 */
extern void lro_receive_frags_compat(struct net_lro_mgr *lro_mgr,
			struct skb_frag_struct *frags, int len, int true_size,
			void *priv, __wsum sum);

extern void lro_vlan_hwaccel_receive_frags_compat(struct net_lro_mgr *lro_mgr,
			struct skb_frag_struct *frags, int len, int true_size,
			struct vlan_group *vgrp, u16 vlan_tag, void *priv,
			__wsum sum);

/* Forward all aggregated SKBs held by lro_mgr to network stack */
extern void lro_flush_all_compat(struct net_lro_mgr *lro_mgr);

extern void lro_flush_pkt(struct net_lro_mgr *lro_mgr, struct iphdr *iph,
			struct tcphdr *tcph);

extern int be_get_frag_header(struct skb_frag_struct *frag, void **mac_hdr,
				void **ip_hdr, void **tcpudp_hdr,
				u64 *hdr_flags, void *priv);
#else

#define	lro_flush_all(m)		do {} while(0)

#endif
